using System;
using HarmonyLib;
using ProtoBuf;
using Vintagestory.API.Client;
using Vintagestory.API.Common;
using Vintagestory.API.Common.Entities;
using Vintagestory.API.MathTools;
using Vintagestory.API.Server;
using Vintagestory.GameContent;

namespace FallDamageOptions {
	public class FallDamageOptionsConfig {
		public bool enabled = true;
		public float max_damage = 0;
		public float min_health = 0;
		public bool speed_based = false;
		public float sb_threshold = -0.2f;
		public float sb_coefficient = 360;
		public bool client_authoritative = false;
	}

	[ProtoContract(ImplicitFields = ImplicitFields.AllPublic)]
	public class FDO_InitPacket {
		public bool authoritative;
	}

	[ProtoContract(ImplicitFields = ImplicitFields.AllPublic)]
	public class FDO_FallPacket {
		public Vec3d positionBeforeFalling;
		public double withYMotion;
	}

	public class FallDamageOptionsSystem: ModSystem {
		public const string configFilename = "falldamageoptions.json";
		public const string channelName = "falldamageoptions";
		protected const int clientTickInterval = 1000;
		protected const string harmonyId = "falldamageoptions";

		protected Harmony harmony;

		// Client.
		public static ICoreClientAPI capi;
		public static IClientNetworkChannel cnc;
		protected long clientListenerId = -1;

		// Server.
		public static FallDamageOptionsConfig config;
		public static IServerNetworkChannel snc;

		public override void Start(ICoreAPI api) {
			base.Start(api);

			this.harmony = new Harmony(harmonyId);
		}

		public override void StartClientSide(ICoreClientAPI capi) {
			base.StartClientSide(capi);
			FallDamageOptionsSystem.capi = capi;

			FallDamageOptionsSystem.cnc = FallDamageOptionsSystem.capi.Network.RegisterChannel(channelName)
					.RegisterMessageType(typeof(FDO_InitPacket))
					.RegisterMessageType(typeof(FDO_FallPacket))
					.SetMessageHandler<FDO_InitPacket>(clientHandleInitPacket);

			this.harmony.Patch(typeof(EntityPlayer).GetMethod(nameof(EntityPlayer.OnFallToGround)), prefix: new HarmonyMethod(typeof(FDO_Client).GetMethod(nameof(FDO_Client.Handle))));

			// Harass server because channel registration does not have a callback for some absurdly inane reason.
			this.clientListenerId = FallDamageOptionsSystem.capi.Event.RegisterGameTickListener(this.clientTick, clientTickInterval);
		}

		protected void clientTick(float obj) {
			// Ask if server is client-authoritative.
			if(FallDamageOptionsSystem.cnc.Connected) {
				FallDamageOptionsSystem.cnc.SendPacket<FDO_InitPacket>(new FDO_InitPacket());
			}
		}

		public override void StartServerSide(ICoreServerAPI sapi) {
			base.StartServerSide(sapi);

			// Load or create config file.
			try {
				FallDamageOptionsSystem.config = sapi.LoadModConfig<FallDamageOptionsConfig>(configFilename);
			} catch(Exception) {}
			if(FallDamageOptionsSystem.config != null) {
				// Save config file in case of missing fields.
				sapi.StoreModConfig<FallDamageOptionsConfig>(FallDamageOptionsSystem.config, configFilename);
			} else {
				// Create new config file.
				FallDamageOptionsSystem.config = new FallDamageOptionsConfig();
				sapi.StoreModConfig<FallDamageOptionsConfig>(FallDamageOptionsSystem.config, configFilename);
			}

			FallDamageOptionsSystem.snc = sapi.Network.RegisterChannel(channelName)
					.RegisterMessageType(typeof(FDO_InitPacket))
					.RegisterMessageType(typeof(FDO_FallPacket))
					.SetMessageHandler<FDO_InitPacket>(serverHandleInitPacket);

			if(FallDamageOptionsSystem.config.client_authoritative) {
				FallDamageOptionsSystem.snc.SetMessageHandler<FDO_FallPacket>(serverHandleFallPacket);

				this.harmony.Patch(typeof(EntityBehaviorHealth).GetMethod(nameof(EntityBehaviorHealth.OnFallToGround)), prefix: new HarmonyMethod(typeof(FDO_Server).GetMethod(nameof(FDO_Server.Ignore))));
			} else {
				this.harmony.Patch(typeof(EntityBehaviorHealth).GetMethod(nameof(EntityBehaviorHealth.OnFallToGround)), prefix: new HarmonyMethod(typeof(FDO_Server).GetMethod(nameof(FDO_Server.Handle))));
			}
		}

		protected void clientHandleInitPacket(FDO_InitPacket packet) {
			FDO_Client.authoritative = packet.authoritative;

			// Stop asking.
			FallDamageOptionsSystem.capi.Event.UnregisterGameTickListener(this.clientListenerId);
			this.clientListenerId = -1;
		}

		protected void serverHandleInitPacket(IServerPlayer fromPlayer, FDO_InitPacket packet) {
			FallDamageOptionsSystem.snc.SendPacket<FDO_InitPacket>(new FDO_InitPacket() {
				authoritative = FallDamageOptionsSystem.config.client_authoritative
			}, new IServerPlayer[] { fromPlayer });
		}

		protected void serverHandleFallPacket(IServerPlayer fromPlayer, FDO_FallPacket packet) {
			FDO_Server.Handle(fromPlayer.Entity.GetBehavior<EntityBehaviorHealth>(),
					packet.positionBeforeFalling,
					packet.withYMotion);
		}

		public override void Dispose() {
			base.Dispose();

			if(this.clientListenerId != -1) {
				FallDamageOptionsSystem.capi.Event.UnregisterGameTickListener(this.clientListenerId);
			}

			if(this.harmony != null) {
				this.harmony.UnpatchAll(harmonyId);
			}
		}
	}

	public class FDO_Client {
		public static bool authoritative = false;

		public static bool Handle(EntityPlayer __instance, double motionY) {
			if(FDO_Client.authoritative
					&& __instance.World.Side == EnumAppSide.Client) {
				// Send fall data to server.
				FallDamageOptionsSystem.cnc.SendPacket<FDO_FallPacket>(new FDO_FallPacket() {
					positionBeforeFalling = __instance.PositionBeforeFalling,
					withYMotion = motionY
				});
			}
			return true;
		}
	}

	public class FDO_Server {
		public static bool Handle(EntityBehaviorHealth __instance, Vec3d positionBeforeFalling, double withYMotion) {
			Entity entity = __instance.entity;

			if(!(entity is EntityPlayer)) {
				return true;
			}

			if(!entity.Properties.FallDamage
					|| !FallDamageOptionsSystem.config.enabled) {
				// Fall damage disabled.
				return false;
			}
			
			bool gliding = (entity as EntityAgent)?.ServerControls.Gliding == true;
			double damage;

			if(FallDamageOptionsSystem.config.speed_based) {
				// Speed-based option.
				if(gliding) {
					withYMotion /= 2;
				}

				if(withYMotion >= FallDamageOptionsSystem.config.sb_threshold) {
					return false;
				}

				damage = FallDamageOptionsSystem.config.sb_coefficient * Math.Pow(withYMotion - FallDamageOptionsSystem.config.sb_threshold, 2);
			} else {
				// Default.
				double yDistance = Math.Abs(positionBeforeFalling.Y - entity.Pos.Y);

				if(yDistance < 3.5f) {
					return false;
				}

				if(gliding) {
					yDistance = Math.Min(14, yDistance);
					withYMotion /= 2;
				}

				if(withYMotion > -0.19) {
					return false;
				}

				damage = yDistance - 3.5f;

				double expectedYMotion = -0.041f * Math.Pow(damage, 0.75f) - 0.22f;
				double yMotionLoss = Math.Max(0, -expectedYMotion + withYMotion);
				damage -= 20 * yMotionLoss;
			}

			if(FallDamageOptionsSystem.config.min_health > 0
					&& damage > __instance.Health) {
				// Non-lethal damage option.
				damage = __instance.Health - FallDamageOptionsSystem.config.min_health;
			}

			if(damage <= 0) {
				// No damage.
				return false;
			}

			if(FallDamageOptionsSystem.config.max_damage > 0
					&& damage > FallDamageOptionsSystem.config.max_damage) {
				// Damage limit option.
				damage = FallDamageOptionsSystem.config.max_damage;
			}

			entity.ReceiveDamage(new DamageSource() {
				Source = EnumDamageSource.Fall,
				Type = EnumDamageType.Gravity
			}, (float)damage);

			// Skip original method.
			return false;
		}

		public static bool Ignore(EntityBehaviorHealth __instance, Vec3d positionBeforeFalling, double withYMotion) {
			// Skip original method.
			return false;
		}
	}
}